__all__ = ("STDCopy",)


import sys
from io import TextIOWrapper, BytesIO
from typing import Literal


_STD = Literal["in", "out", "err", "stdin", "stdout", "stderr"]


class STDCopy(TextIOWrapper):
    """
    Captures ``sys.std*``.

    Usage
    -----
    You can either set the ``sys.std*``-attribute to this class,
    or you can use it inside a ``with``-statements, which will
    then capture just inside the statement.

    Attributes
    ----------
    captured: str
        Everything which got captured.

    Notes
    -----
    ``stdin`` (or ``in``) doesn't work at the moment and will raise an ``EOFError``.
    """

    captured: str
    _std: _STD
    _sys_std: TextIOWrapper

    __slots__ = (
        "captured",
        "_std",
        "_sys_std",
    )

    def __init__(self, std: _STD, *args, **kwargs):
        """
        Parameters
        ----------
        std: _STD
            The sys.std* to capture.
        args, kwargs: ...
            args and kwargs for TextIOWrapper
        """
        if not std.startswith("std"):
            std = "std" + std
        self._std = std
        self._sys_std = getattr(sys, std.lower())
        self.captured = ""

        if std == "stdin":
            import warnings

            warnings.warn(
                "Capturing sys.stdin is currently unstable and might not work!",
                RuntimeWarning,
            )

        super().__init__(BytesIO(), *args, **kwargs)

    def write(self, s):
        ret = self._sys_std.write(s)
        self.captured += s
        return ret

    def read(self, size=-1):
        ret = self._sys_std.read(size)
        self.captured += ret
        return ret.removesuffix("\n")

    def __enter__(self):
        setattr(sys, self._std.lower(), self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        setattr(sys, self._std.lower(), self._sys_std)



__all__ = ("get_logger",)


import sys
import typing
from logging import (
    getLogger,
    Formatter,
    StreamHandler,
)


_LOG_LEVEL_STR = typing.Literal[
    "CRITICAL",
    "FATAL",
    "ERROR",
    "WARN",
    "WARNING",
    "INFO",
    "DEBUG",
    "NOTSET",
]
_F = "\033[33m{asctime} \t{name: <15} {levelname: <10}\t{message}\033[0m"
_logging_formatter = Formatter(_F, style="{")
_logging_handler = StreamHandler(sys.stdout)
_logging_handler.setFormatter(_logging_formatter)


def get_logger(
    name: typing.Optional[str],
    *,
    level: typing.Union[_LOG_LEVEL_STR, int, None] = "DEBUG",
    add_handler: bool = True,
):
    """
    Parameters
    ----------
    name: str, optional
        The name from the logger.
        (`root` if `None`)
    level: _LOG_LEVEL_STR, int, optional
        The loglevel.
    add_handler: bool
        Whether a handler should be added or not.

    Returns
    -------
    logging.Logger
    """
    logger = getLogger(name)
    if add_handler:
        logger.addHandler(_logging_handler)
    if level is not None:
        if isinstance(level, str):
            level = level.upper()
        logger.setLevel(level)
    return logger



try:
    from .ratelimit import *

    # insert other imports here...
except ImportError:
    import warnings

    warnings.warn(
        "{pkg!r} must be installed via 'pip install AlbertUnruhUtils.py[{extra}]'".format(
            pkg=__package__, extra="async"
        ),
        category=UserWarning,
    )

    raise



__all__ = (
    "deprecated",
    "copy_docs",
    "not_implemented",
)


import re
import typing
import warnings
from copy import copy
from functools import wraps


_Version = typing.Union[str, tuple[int]]
_Function = typing.TypeVar("_Function", bound=typing.Callable)
_Docs_NumPy = typing.Literal["NumPy", "numpy", "NumPyDoc", "numpydoc"]
_Doc_Style = typing.Union[_Docs_NumPy]


def deprecated(
    since: _Version = None,
    *,
    instead: typing.Union[str, typing.Callable] = None,
    update_docs: bool = True,
    doc_style: _Doc_Style = "NumPy",
) -> typing.Callable[[_Function], _Function]:
    """
    Marks a function/method as deprecated.

    Parameters
    ----------
    since: _Version
        Since when the function/method is deprecated.
    instead: str, typing.Callable, optional
        What function/method should be used instead.
    update_docs: bool
        Whether the docs should be updated or not.
    doc_style: _Doc_Style
        The DocStyle to use to update the docs.
    """

    def outer(func: _Function) -> _Function:
        message = "{0.__name__} is deprecated".format(func)
        if since is not None:
            if not isinstance(since, str):
                version = ".".join(str(v) for v in since)
            else:
                version = since
            message += " since version {0}".format(version)
        message += "."
        if instead is not None:
            if not isinstance(instead, str):
                name = instead.__name__
            else:
                name = instead
            message += " Use {0} instead.".format(name)

        if update_docs:
            # know issue: if a class isn't initialized
            # the method will be called function
            f_o_c = func.__class__.__name__

            # NumPy
            if doc_style in _Docs_NumPy.__args__:  # type: ignore
                lines = [
                    f"",
                    f"Deprecation",
                    f"-----------",
                    f"This {f_o_c} " + message.split(maxsplit=1)[-1],
                    f"",
                ]

            # invalid
            else:
                supported = []
                for g in _Doc_Style.__args__:  # type: ignore
                    for s in g.__args__:
                        supported.append(s)
                raise ValueError(
                    f"Unknown or unsupported formatting style {doc_style!r} for docstrings! "
                    f"Use one of them instead: {', '.join(supported)}"
                )

            tab = ""
            if func.__doc__:
                res = re.findall(r"^\s+", func.__doc__)
                if res:
                    tab = res[0].removeprefix("\n")
            func.__doc__ += "\n".join(tab + line for line in lines)

        @wraps(func)
        def inner(*args, **kwargs):
            filters = copy(warnings.filters)
            warnings.filters = []
            warnings.warn(message=message, category=DeprecationWarning)
            warnings.filters = filters

            return func(*args, **kwargs)

        return inner

    return outer


def not_implemented(
    reason: str = None,
    *,
    update_docs: bool = True,
    doc_style: _Doc_Style = "NumPy",
) -> typing.Callable[[_Function], _Function]:
    """
    Marks a function/method as not implemented, but as coming soon.
    (And they might open a PR to implement it for you :D)

    Parameters
    ----------
    reason: str
        The reason why this function/method is not implemented yet.
    update_docs: bool
        Whether the docs should be updated or not.
    doc_style: _Doc_Style
        The DocStyle to use to update the docs.
    """

    def outer(func: _Function) -> _Function:
        message = "{0.__name__} is not implemented yet".format(func)
        if reason:
            message += " with following reason: {0}".format(reason)

        if not message.endswith("."):
            message += "."

        if update_docs:
            # know issue: if a class isn't initialized
            # the method will be called function
            f_o_c = func.__class__.__name__

            # NumPy
            if doc_style in _Docs_NumPy.__args__:  # type: ignore
                # this goes on top of the __doc__
                lines0 = [
                    f"",
                    f"This {f_o_c} " + message.split(maxsplit=1)[-1],
                    f"",
                ]
                # this goes at the end of the __doc__
                lines1 = [
                    f"",
                    f"Raises",
                    f"------",
                    f"NotImplementedError",
                    f"",
                ]
                if reason:
                    lines1.insert(-1, " " * 4 + reason)

            # invalid
            else:
                supported = []
                for g in _Doc_Style.__args__:  # type: ignore
                    for s in g.__args__:
                        supported.append(s)
                raise ValueError(
                    f"Unknown or unsupported formatting style {doc_style!r} for docstrings! "
                    f"Use one of them instead: {', '.join(supported)}"
                )

            tab = ""
            if func.__doc__:
                res = re.findall(r"^\s+", func.__doc__)
                if res:
                    tab = res[0].removeprefix("\n")
            else:
                func.__doc__ = ""
            doc_start = "\n" if func.__doc__.startswith("\n") else ""
            func.__doc__ = (
                doc_start
                + ("\n".join(tab + line for line in lines0))
                + (func.__doc__ or "")
            )
            func.__doc__ += "\n".join(tab + line for line in lines1)

        @wraps(func)
        def inner(*_, **__):
            filters = copy(warnings.filters)
            warnings.filters = []
            warnings.warn(message=message, category=UserWarning)
            warnings.filters = filters

            raise NotImplementedError(message)

        return inner

    return outer


def copy_docs(
    docs: typing.Union[str, object],
    *,
    append: bool = False,
    add_copy_note: bool = False,
    custom_origin: str = None,
) -> typing.Callable[[_Function], _Function]:
    """
    Copies docs from an object or str to the decorated function/method/class.

    Parameters
    ----------
    docs: str, object
        The docs to copy.
    append: bool
        Whether the docs should be appended or be replaced.
    add_copy_note: bool
        Whether the docs should have a note that it was copied or not.
    custom_origin: str, optional
        The origin which should be displayed if ``add_copy_note`` is set to ``True``.
    """
    _docs = copy(docs)
    if not isinstance(docs, str):
        docs = docs.__doc__

    if add_copy_note:
        if custom_origin:
            origin = custom_origin
        else:
            try:
                origin = _docs.__name__
            except AttributeError:
                origin = _docs.__class__.__name__
        lines = [
            f"",
            f"",
            f"copied from {origin}",
            f"",
        ]
    else:
        lines = []

    def decorator(func: _Function) -> _Function:
        # to keep the indentation-style
        tab = ""
        if func.__doc__:
            res = re.findall(r"^\s+", func.__doc__)
            if res:
                tab = res[0].removeprefix("\n")

        if not append:
            func.__doc__ = ""

        # to be able to replace it later with ``tab``
        docs_tab = ""
        if docs:
            res = re.findall(r"^\s+", docs)
            if res:
                docs_tab = res[0].removeprefix("\n")

        func.__doc__ += "\n".join(
            tab + line.removeprefix(docs_tab) for line in docs.splitlines()
        )
        func.__doc__ += "\n".join(tab + line for line in lines)

        return func

    return decorator



__all__ = (
    "get_version",
    "is_version",
    "is_lower_version",
    "is_higher_version",
)


import importlib
import re
import types
import typing


def _version_to_tuple(
    version: str,
    /,
    *,
    sep: str = ".",
    keep_dev_version: bool = True,
    semver: bool = False,
) -> tuple[int, ...]:
    """
    Parameters
    ----------
    version: str
        The raw version.
    sep: str
        The separator for e.g. major and minor.
    keep_dev_version: bool
        Whether to keep developer-versions or not.
    semver: bool
        Whether the return should be structured like a semantic version (https://semver.org)

    Returns
    -------
    tuple[int, ...]

    Notes
    -----
    If ``keep_dev_version`` is set to ``True`` the return may also contain ``str``.
    """
    t_version = ()
    for v in version.split(sep):
        try:
            t_version += (int(v),)
        except ValueError:
            # eg "0a3"
            first = re.findall(r"^[\d]+", v)
            if first:
                t_version += (int(first[0]),)

            if keep_dev_version:
                covered = len(sep.join(str(v) for v in t_version))
                t_version += (version[covered:].removeprefix(sep),)
                break

    if semver:
        while len(t_version) < 3:
            t_version += (0,)
        if len(t_version) > 3:
            t_version = t_version[:3] + tuple([sep.join(str(v) for v in t_version[3:])])

    return t_version


def get_version(
    module_o_name: typing.Union[types.ModuleType, str],
    /,
    *,
    sep: str = ".",
    keep_dev_version: bool = True,
    semver: bool = False,
) -> tuple[int, ...]:
    """
    Parameters
    ----------
    module_o_name: types.ModuleType, str
        The module of which you want the version.
    sep: str
        The separator for e.g. major and minor.
    keep_dev_version: bool
        Whether to keep developer-versions or not.
    semver: bool
        Whether the return should be structured like a semantic version (https://semver.org)

    Returns
    -------
    tuple[int, ...]

    Notes
    -----
    If ``keep_dev_version`` is set to ``True`` the return may also contain ``str``.
    """
    # get module
    if isinstance(module_o_name, types.ModuleType):
        module = module_o_name
    elif module_o_name in globals():
        module = globals()[module_o_name]
    else:
        module = importlib.import_module(module_o_name)

    # get version
    if hasattr(module, "version_info"):
        version = sep.join(str(v) for v in module.version_info)
    elif hasattr(module, "__version__"):
        version = module.__version__
    elif hasattr(module, "version"):
        version = module.version
    else:
        version = "0.0.0a"  # one of the lowest versions if no version is set

    return _version_to_tuple(
        version,
        sep=sep,
        keep_dev_version=keep_dev_version,
        semver=semver,
    )


def is_version(
    module_o_version: typing.Union[types.ModuleType, str, tuple[int, ...]],
    version: typing.Union[tuple[int, ...], str],
    /,
    *,
    sep: str = ".",
    keep_dev_version: bool = True,
    semver: bool = False,
) -> bool:
    """
    MODULE_o_VERSION == VERSION

    Parameters
    ----------
    module_o_version: types.ModuleType, str, tuple[int, ...]
        The module or version of which you want to check the version.
    version: tuple[int, ...], str
        The version to check against.
    sep: str
        The separator for e.g. major and minor.
    keep_dev_version: bool
        Whether to keep developer-versions or not.
    semver: bool
        Whether the return should be structured like a semantic version (https://semver.org)

    Returns
    -------
    bool
    """
    try:
        # module?
        m_version = get_version(
            module_o_version,
            sep=sep,
            keep_dev_version=keep_dev_version,
            semver=semver,
        )
    except ModuleNotFoundError:
        # version!
        if isinstance(module_o_version, str):
            m_version = _version_to_tuple(
                module_o_version,
                sep=sep,
                keep_dev_version=keep_dev_version,
                semver=semver,
            )
        else:
            m_version = module_o_version

    if isinstance(version, str):
        version = _version_to_tuple(
            version,
            sep=sep,
            keep_dev_version=keep_dev_version,
            semver=semver,
        )

    return m_version == version


def is_lower_version(
    module_o_version: typing.Union[types.ModuleType, str, tuple[int, ...]],
    version: typing.Union[tuple[int, ...], str],
    /,
    *,
    sep: str = ".",
    keep_dev_version: bool = True,
    semver: bool = False,
) -> bool:
    """
    MODULE_o_VERSION < VERSION

    Parameters
    ----------
    module_o_version: types.ModuleType, str, tuple[int, ...]
        The module or version of which you want to check the version.
    version: tuple[int, ...], str
        The version to check against.
    sep: str
        The separator for e.g. major and minor.
    keep_dev_version: bool
        Whether to keep developer-versions or not.
    semver: bool
        Whether the return should be structured like a semantic version (https://semver.org)

    Returns
    -------
    bool
    """
    try:
        # module?
        m_version = get_version(
            module_o_version,
            sep=sep,
            keep_dev_version=keep_dev_version,
            semver=semver,
        )
    except ModuleNotFoundError:
        # version!
        if isinstance(module_o_version, str):
            m_version = _version_to_tuple(
                module_o_version,
                sep=sep,
                keep_dev_version=keep_dev_version,
                semver=semver,
            )
        else:
            m_version = module_o_version

    if isinstance(version, str):
        version = _version_to_tuple(
            version,
            sep=sep,
            keep_dev_version=keep_dev_version,
            semver=semver,
        )

    # workaround for dev-versions
    if len(m_version) != len(version):
        for m, v in zip(m_version, version):
            if m != v:
                break
        else:
            # one of these have e.g. "alpha" at the end
            return len(m_version) > len(version)

    return m_version < version


def is_higher_version(
    module_o_version: typing.Union[types.ModuleType, str, tuple[int, ...]],
    version: typing.Union[tuple[int, ...], str],
    /,
    *,
    sep: str = ".",
    keep_dev_version: bool = True,
    semver: bool = False,
) -> bool:
    """
    MODULE_o_VERSION > VERSION

    Parameters
    ----------
    module_o_version: types.ModuleType, str, tuple[int, ...]
        The module or version of which you want to check the version.
    version: tuple[int, ...], str
        The version to check against.
    sep: str
        The separator for e.g. major and minor.
    keep_dev_version: bool
        Whether to keep developer-versions or not.
    semver: bool
        Whether the return should be structured like a semantic version (https://semver.org)

    Returns
    -------
    bool
    """
    args = [
        module_o_version,
        version,
    ]
    kwargs = {
        "sep": sep,
        "keep_dev_version": keep_dev_version,
        "semver": semver,
    }

    lt = is_lower_version(*args, **kwargs)  # noqa
    eq = is_version(*args, **kwargs)  # noqa

    return not (lt or eq)



__all__ = ("TeX",)


import typing
from pathlib import Path
from io import BytesIO
from os import PathLike

import matplotlib.pyplot as plt
from PIL import Image, ImageChops

from ..utils import not_implemented
from .. import __url__


_PathLike = typing.Union[
    PathLike,
    Path,
    str,
]
_Color = typing.Union[
    typing.Iterable[float],
    str,
]


class TeX:
    _color: _Color
    _file: Path
    _format: str
    _tex: str

    __slots__ = (
        "_color",
        "_file",
        "_format",
        "_tex",
    )

    def __init__(
        self,
        tex: str,
        *,
        file: _PathLike = "tex.png",
        format: str = "png",  # noqa
        color: _Color = "#fe4b03",  # aka "blood orange"
    ) -> None:
        """
        Parameters
        ----------
        tex: str
            The input which should be displayed.
        file: _PathLike
            Sets the default value for `file`.
        format
            Sets the default value for `format`.
        color
            Sets the default value for `color`.
        """
        if not tex.startswith("$"):
            tex = f"${tex}$"

        self._tex = tex
        self._file = Path(file)
        self._format = format
        self._color = color

    @classmethod
    @not_implemented(
        f"This feature is coming soon. Feel free to push it and open a PR on GitHub ({__url__})."
    )
    def from_python_code(
        cls,
        function: typing.Callable,
        *,
        file: _PathLike = "tex.png",
        format: str = "png",  # noqa
        color: _Color = "#fe4b03",  # aka "blood orange"
    ) -> "TeX":
        """
        Creates TeX from a function.

        Parameters
        ----------
        function: typing.Callable
            The Function which should be converted to TeX.
        file: _PathLike
            Sets the default value for `file`.
        format
            Sets the default value for `format`.
        color
            Sets the default value for `color`.

        Returns
        -------
        TeX
        """

    @property
    def tex(self) -> str:
        return self._tex

    @property
    def default_color(self) -> _Color:
        return self._color

    @property
    def default_file(self) -> Path:
        return self._file

    @property
    def default_format(self) -> str:
        return self._format

    def create_image(
        self,
        *,
        format: typing.Optional[str] = None,  # noqa
        color: typing.Optional[_Color] = None,
    ) -> Image.Image:
        """
        Creates the TeX-image.

        Parameters
        ----------
        format: str, optional
            If `None` the default for `format` 'll be used.
        color: _Color, optional
            If `None` the default for `color` 'll be used.

        Returns
        -------
        Image.Image
            The TeX-image.
        """
        if format is None:
            format = self.default_format  # noqa
        if color is None:
            color = self.default_color

        buffer = BytesIO()
        plt.rc("text", usetex=True)
        plt.axis("off")
        plt.text(0, 0, self._tex, size=40, color=color)
        plt.savefig(buffer, format=format, transparent=True)
        plt.close()

        image = Image.open(buffer)
        bg = Image.new(image.mode, image.size, (0,) * 4)  # type: ignore
        diff = ImageChops.difference(image, bg)
        bbox = diff.getbbox()
        return image.crop(bbox)

    def save_to_file(
        self,
        file: typing.Optional[_PathLike] = None,
        /,
        format: typing.Optional[str] = None,  # noqa
        color: typing.Optional[_Color] = None,
    ) -> Path:
        """
        Saves the TeX-image to a file.

        Parameters
        ----------
        file: _PathLike, optional
            If `None` the default for `file` 'll be used.
        format: str, optional
            If `None` the default for `format` 'll be used.
        color: _Color, optional
            If `None` the default for `color` 'll be used.

        Returns
        -------
        Path
            The path to the saved TeX-image.
        """
        if file is None:
            file = self.default_file
        else:
            file = Path(file)
        if format is None:
            format = self.default_format  # noqa
        if color is None:
            color = self.default_color

        image = self.create_image(format=format, color=color)
        image.save(file)

        return file


def __main():
    """
    This is just a little function to test our TeX-cLaSs.
    """
    eulers_identity = r"e^{i\pi}+1=0"
    file = TeX(eulers_identity).save_to_file()
    print(f"saved Euler's Identity to {file.absolute()}")



__all__ = ("ServerRateLimit",)


import functools
import typing
import uuid
from aioredis import Redis
from time import time


C_IN = typing.TypeVar("C_IN")
C_OUT = typing.TypeVar("C_OUT")


class ServerRateLimit:
    """Docs 'll come soon... (If you want docs right now you can take a look into ``__init__``)"""

    sections: dict[str, dict[str, int]]
    retrieve_section: typing.Callable[[...], typing.Awaitable[tuple[str, str]]]

    __slots__ = ("sections", "retrieve_section", "_redis")

    def __init__(
        self,
        sections: dict[str, dict[str, int]],
        retrieve_section: typing.Callable[
            [...], typing.Awaitable[tuple[str, typing.Union[str, int]]]
        ],
        *,
        redis: Redis = None,
    ):
        """
        Parameters
        ----------
        sections: dict[str, dict[str, int]]
            Parameter ``sections`` requires following structure:
            ```py
            >>> {
            ...     "<NAME or TYPE (e.g. user, admin etc.)>": {
            ...         # type: int
            ...         "amount": 10
            ...
            ...         # type: int
            ...         "interval": 60  # in seconds
            ...
            ...         # type: int
            ...         "timeout": 60  # in seconds  # if a section requests to often then the timeout 'll be applied
            ...     },
            ...     "<second NAME or TYPE>": {
            ...         ...
            ...     },
            ...     ...
            ... }
            ```
        retrieve_section: typing.Callable[[...], typing.Awaitable[tuple[str, typing.Union[str, int]]]]
            This function 'll feed all it's data from the original callable.
            e.g. ```py
            >>> @ServerRateLimit({"user": {...}, "admin": {...}}, retrieve)
            ... async def foo(*args, **kwargs) -> ...:
            ...     pass
            ...
            >>> async def retrieve(*args, **kwargs) -> (str, str):
            ...     '''This is just an example, you have to manage yourself how you
            ...        set it (can also be static by using a simple lambda-expression)'''
            ...     if "admin_id" in kwargs:
            ...         return "admin", 0
            ...     return "user", 0
            ```
        redis: Redis, optional
            An own redis can optionally be set.

        Notes
        -----
        The first return value from ``retrieve_section``
        is the ``section``, the second is the ``id`` to
        have every section separated.
        """
        self.sections = sections
        self.retrieve_section = retrieve_section

        if redis is None:
            redis = Redis(host="127.0.0.1", port=6262, db=0)
        self._redis = redis

    def __call__(
        self,
        func: typing.Callable[[C_IN], typing.Awaitable[C_OUT]],
    ) -> typing.Callable[
        [C_IN], typing.Awaitable[tuple[tuple[bool, dict[str, int]], C_OUT]]
    ]:
        async def decorator(
            *args, **kwargs
        ) -> tuple[tuple[bool, dict[str, int]], C_OUT]:
            """
            Returns
            -------
            tuple[tuple[bool, dict[str, int]], C_OUT]
            """
            section, id = await self.retrieve_section(*args, **kwargs)  # noqa
            if section not in self.sections:
                raise RuntimeError(
                    "Can't use key {section!r}. You have to return one of the following: {possible}".format(
                        section=section,
                        possible=", ".join(f"{k!r}" for k in self.sections),
                    )
                )

            await self._check_timeout(section, id)

            timeout = await self._calculate_timeout(section, id)
            remaining = await self._calculate_remaining_calls(section, id)

            data = {
                "request": {
                    "remaining": -1,
                    "limit": self.sections[section]["amount"],
                    "period": self.sections[section]["interval"],
                    "timeout": timeout,
                }
            }

            if not remaining > 0 or timeout:
                data["request"]["remaining"] = await self._calculate_remaining_calls(
                    section, id
                )
                return (False, data), ()

            await self._record_call(section, id)
            data["request"]["remaining"] = await self._calculate_remaining_calls(
                section, id
            )
            return (True, data), await func(*args, **kwargs)

        return functools.update_wrapper(decorator, func)

    async def _record_call(
        self,
        section: str,
        id: typing.Union[str, int],  # noqa
    ) -> None:
        """
        Parameters
        ----------
        section: str
        id: str, int
        """
        key = f"call-{section}-{id}"
        await self._redis.execute_command(
            f"ZADD {key} {time()+self.sections[section]['interval']} {uuid.uuid4()}"
        )
        await self._redis.expire(key, self.sections[section]["interval"])

    async def _calculate_remaining_calls(
        self,
        section: str,
        id: typing.Union[str, int],  # noqa
    ) -> int:
        """
        Parameters
        ----------
        section: str
        id: str, int

        Returns
        -------
        int
        """
        key = f"call-{section}-{id}"

        # cleanup
        await self._redis.zremrangebyscore(key, 0, time())

        return self.sections[section]["amount"] - int(
            await self._redis.zcount(key, 0, 2**62) or 0
        )

    async def _check_timeout(
        self,
        section: str,
        id: typing.Union[str, int],  # noqa
    ) -> None:
        """
        Parameters
        ----------
        section: str
        id: str, int
        """
        key = f"cooldown-{section}-{id}"

        if not await self._calculate_remaining_calls(section, id) > 0:
            if not await self._redis.exists(key):
                await self._redis.append(key, 1)
                await self._redis.expire(key, self.sections[section]["timeout"])

    async def _calculate_timeout(
        self,
        section: str,
        id: typing.Union[str, int],  # noqa
    ) -> int:
        """
        Parameters
        ----------
        section: str
        id: str, int

        Returns
        -------
        int
        """
        key = f"cooldown-{section}-{id}"

        return max(0, await self._redis.ttl(key))


if __name__ == "__main__":
    r = Redis()

    async def _():
        return "default", 0  # noqa

    @ServerRateLimit(
        {"default": {"interval": 10, "amount": 10, "timeout": 20}},
        _,
        redis=r,
    )
    async def test():
        return "<--test()-->"

    async def main():
        from asyncio import sleep

        while True:
            print(test.__name__, await test())
            await sleep(0.5)

    from asyncio import run

    run(main())



from setuptools import setup
import re


with open("./AlbertUnruhUtils/__init__.py") as f:
    file = f.read()

version = re.search(
    r"^__version__\s*=\s*[\'\"]([^\'\"]*)[\'\"]", file, re.MULTILINE
).group(1)
url = re.search(r"^__url__\s*=\s*[\'\"]([^\'\"]*)[\'\"]", file, re.MULTILINE).group(1)
license = re.search(  # noqa
    r"^__license__\s*=\s*[\'\"]([^\'\"]*)[\'\"]", file, re.MULTILINE
).group(1)
author = re.search(
    r"^__author__\s*=\s*[\'\"]([^\'\"]*)[\'\"]", file, re.MULTILINE
).group(1)
description = re.search(
    r"^__description__\s*=\s*[\'\"]([^\'\"]*)[\'\"]", file, re.MULTILINE
).group(1)


# not in setup.cfg for GitHub's Dependency-Graph
install_requires = [
    "redis~=4.1.0",
    "pillow~=9.0.0",
    "matplotlib~=3.5.1",
]
extras_require = {
    "async": [
        "aioredis~=2.0.1",
    ],
}


# not really my code... (https://github.com/Rapptz/discord.py/blob/master/setup.py#L15)
if version[-1].isalpha():
    # append version identifier based on commit count
    try:
        import subprocess

        p = subprocess.Popen(
            ["git", "rev-list", "--count", "HEAD"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        if out:
            version += out.decode("utf-8").strip()

        p = subprocess.Popen(
            ["git", "rev-parse", "--short", "HEAD"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        if out:
            version += "+g" + out.decode("utf-8").strip()

    except Exception as e:
        from warnings import warn

        warn(
            message=f"\033[31mUnable to append version identifier!\033[0m "
            f"\033[35m{e.__class__.__name__}: {e}\033[0m",
            category=UserWarning,
        )

setup(
    version=version,
    url=url,
    license=license,
    author=author,
    description=description,
    install_requires=install_requires,
    extras_require=extras_require,
)



__author__ = "AlbertUnruh"
__url__ = "https://github.com/AlbertUnruh/AlbertUnruhUtils.py"
__license__ = "MIT"
__copyright__ = f"(c) 2021 - present {__author__}"
__version__ = "2.2.0"
__description__ = "A collection of utils written in Python."

from . import config
from . import ratelimit
from . import utils
from . import visual

from .config import *
from .ratelimit import *
from .utils import *
from .visual import *



__all__ = ("JSONConfig",)


import typing
from json import (
    dump,
    load,
)


DEFAULT_CONFIG = {
    "message": "automatically generated configuration (this entry can be deleted)"
}


class JSONConfig:
    """Docs 'll come soon..."""

    __slots__ = (
        "default",
        "_config",
        "_file",
        "_default_config",
    )

    def __init__(
        self,
        *,
        file: str,
        default_return: typing.Any = None,
        default_config: dict = None,
    ):
        """
        Parameters
        ----------
        file: str
        default_return: Any
            The default when calling `__getitem__`
        default_config: dict
        """
        try:
            with open(file) as f:
                self._config = load(f)
        except (OSError, ValueError) as e:
            import sys

            print(
                f"Ignoring {e.__class__.__name__} ({'; '.join(str(arg) for arg in e.args)}); "
                f"Overwriting (existing) configuration at {file!r}",
                file=sys.stderr,
            )

            self._config = default_config or DEFAULT_CONFIG
            with open(file, "w") as f:
                dump(self._config, f, indent=4)

        self.default = default_return
        self._file = file
        self._default_config = default_config

    @property
    def file(self) -> str:
        return self._file

    @file.setter
    def file(self, value: str) -> None:
        self.__init__(
            file=value, default_return=self.default, default_config=self._default_config
        )

    @property
    def config(self) -> dict:
        return self._config

    @config.setter
    def config(self, value: str) -> None:
        assert isinstance(
            value, dict
        ), f"{self.__class__.__name__}.config must be an instance of 'dict', not {value.__class__.__name__!r}!"
        self._config = value
        with open(self._file, "w") as f:
            dump(self._config, f, indent=4)

    def __getitem__(self, item):
        return self._config.get(item, self.default)

    def __setitem__(self, key, value):
        self._config[key] = value
        with open(self._file, "w") as f:
            dump(self._config, f, indent=4)
